/*
 * $Id: sst.c,v 1.1.1.1 2006/10/16 15:47:29 danov Exp $
 *
 * Flash driver for SST SST39VF020
 * Copyright (C) 2003 AH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Written by August Hörandl <august.hoerandl@gmx.at>
 * Modified by Marcel Telka <marcel@telka.sk>, 2003.
 *
 * Documentation:
 * [1] Advanced Micro Devices, "Am29LV640D/Am29LV641D",
 *     September 20, 2002     Rev B, 22366b8.pdf
 * [2] Advanced Micro Devices, "Am29LV642D",
 *     August 14, 2001    Rev A, 25022.pdf
 *
 */

#include "sysdep.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <flash/cfi.h>
#include <flash/intel.h>
#include <unistd.h>

#include <brux/flash.h>
#include <brux/bus.h>
#include <brux/cfi.h>

static int dbg = 0;

static int sst_flash_erase_block( cfi_array_t *cfi_array, uint32_t adr );
static int sst_flash_unlock_block( cfi_array_t *cfi_array, uint32_t adr );
static int sst_flash_program( cfi_array_t *cfi_array, uint32_t adr, uint32_t data );
static void sst_flash_read_array( cfi_array_t *cfi_array ); 

static int o;

/* autodetect, we can handle this chip */
static int 
sst_flash_autodetect32( cfi_array_t *cfi_array )
{
	if(cfi_array->bus_width != 4) return 0;
	o = 2; /* Heuristic */
	return (cfi_array->cfi_chips[0]->cfi.identification_string.pri_id_code == CFI_VENDOR_SST_SCS);
}

static int 
sst_flash_autodetect16( cfi_array_t *cfi_array )
{
	if(cfi_array->bus_width != 2) return 0;
	o = 1; /* Heuristic */
	return (cfi_array->cfi_chips[0]->cfi.identification_string.pri_id_code == CFI_VENDOR_SST_SCS);
}

static int 
sst_flash_autodetect8( cfi_array_t *cfi_array )
{
	printf("sst_flash_autodetect8...bus_width: %d..\n", cfi_array->bus_width);
	
	if(cfi_array->bus_width != 1) return 0;
	o = 0; /* Heuristic */
	if (cfi_array->cfi_chips[0]->cfi.identification_string.pri_id_code == CFI_VENDOR_SST_SCS)
		printf("Right SST code...\n");
	return (cfi_array->cfi_chips[0]->cfi.identification_string.pri_id_code == CFI_VENDOR_SST_SCS);
}

/*
 * second implementation: see [1], page 30
 */
static int
sststatus( bus_t *bus, uint32_t adr, int data )
{
	int timeout;
	uint32_t togglemask = ((1 << 6) << 16) + (1 << 6); /* DQ 6 */
	/*  int dq5mask = ((1 << 5) << 16) + (1 << 5); DQ5 */

	for (timeout = 0; timeout < 100; timeout++) {
		uint32_t data1 = bus_read( bus, adr );
		uint32_t data2 = bus_read( bus, adr );

		/*printf("sststatus %d: %04X/%04X   %04X/%04X \n", */
		/*	   timeout, data1, data2, (data1 & togglemask), (data2 & togglemask)); */
		if ( (data1 & togglemask) == (data2 & togglemask)) /* no toggle */
			return 1;

		/*    if ( (data1 & dq5mask) != 0 )   TODO */
		/*      return 0; */
		if (dbg) 
			printf( "sststatus %d: %04X/%04X\n", timeout, data1, data2 );
		else
			printf( "." );
		usleep( 100 );
	}
	return 0;
}

static void
sst_flash_print_info( cfi_array_t *cfi_array )
{
	int mid, cid, prot = 0;
	bus_t *bus = cfi_array->bus;

	bus_write( bus, 0x5555 << o, 0x00aa );	/* autoselect p29 */
	bus_write( bus, 0x2aaa << o, 0x0055 );
	bus_write( bus, 0x5555 << o, 0x0090 );
	mid = bus_read( bus, 0x00 << o ) & 0xFFFF;
	cid = bus_read( bus, 0x01 << o ) & 0xFFFF;
//	prot = bus_read( bus, 0x02 << o ) & 0xFF;
	sst_flash_read_array( cfi_array );		/* SST reset */
	printf( _("Chip: SST Flash\n\tManufacturer: ") );
	switch (mid) {
		case 0x00bf:
			printf( _("SST") );
			break;
		default:
			printf( _("Unknown manufacturer (ID 0x%04x)"), mid );
			break;
	}
	printf( _("\n\tChip: ") );
	switch (cid) {
		case 0x00d4:
			printf( _("SST39VF512") );
			break;
		case 0x00d5:
			printf( _("SST39VF010") );
			break;
		case 0x00d6:
			printf( _("SST39VF020") );
			break;
		case 0x00d7:
			printf( _("SST39VF040") );
			break;
		default:
			printf ( _("Unknown (ID 0x%04x)"), cid );
			break;
	}
	printf( _("\n\tProtected: %04x\n"), prot );
}

static int
sst_flash_erase_block( cfi_array_t *cfi_array, uint32_t adr )
{
	bus_t *bus = cfi_array->bus;

	printf("sst_flash_erase_block 0x%08X\n", adr);

	/*	printf("protected: %d\n", sstisprotected(ps, adr)); */

	bus_write( bus, 0x5555 << o, 0x00aa ); /* autoselect p29, sector erase */
	bus_write( bus, 0x2aaa << o, 0x0055 );
	bus_write( bus, 0x5555 << o, 0x0080 );
	bus_write( bus, 0x5555 << o, 0x00aa );
	bus_write( bus, 0x2aaa << o, 0x0055 );
	bus_write( bus, adr, 0x0030 );

	if (sststatus( bus, adr, 0xffff )) {
		printf( "flash_erase_block 0x%08X DONE\n", adr );
		sst_flash_read_array( cfi_array );	/* SST reset */
		return 0;
	}
	printf( "flash_erase_block 0x%08X FAILED\n", adr );
	/* Read Array */
	sst_flash_read_array( cfi_array );		/* SST reset */

	return FLASH_ERROR_UNKNOWN;
}

static int
sst_flash_unlock_block( cfi_array_t *cfi_array, uint32_t adr )
{
	printf( "flash_unlock_block 0x%08X IGNORE\n", adr );
	return 0;
}

static int
sst_flash_program( cfi_array_t *cfi_array, uint32_t adr, uint32_t data )
{
	int status;
	bus_t *bus = cfi_array->bus;

	if (dbg)
		printf("\nflash_program 0x%08X = 0x%08X\n", adr, data);

	bus_write( bus, 0x5555 << o, 0x00aa ); /* autoselect p29, program */
	bus_write( bus, 0x2aaa << o, 0x0055 );
	bus_write( bus, 0x5555 << o, 0x00a0 );

	bus_write( bus, adr, data );
	status = sststatus( bus, adr, data );
	/*	sst_flash_read_array(ps); */

	return !status;
}

static void
sst_flash_read_array( cfi_array_t *cfi_array )
{
	/* Read Array */
	bus_write( cfi_array->bus, 0x0, 0x00f0 ); /* SST reset */
}

flash_driver_t sst_32_flash_driver = {
	4, /* buswidth */
	N_("SST Standard Command Set"),
	N_("supported: SST xxxx; 2x16 Bit"),
	sst_flash_autodetect32,
	sst_flash_print_info,
	sst_flash_erase_block,
	sst_flash_unlock_block,
	sst_flash_program,
	sst_flash_read_array,
};

flash_driver_t sst_16_flash_driver = {
	2, /* buswidth */
	N_("SST Standard Command Set"),
	N_("supported: SST xxxx; 1x16 Bit"),
	sst_flash_autodetect16,
	sst_flash_print_info,
	sst_flash_erase_block,
	sst_flash_unlock_block,
	sst_flash_program,
	sst_flash_read_array,
};

flash_driver_t sst_8_flash_driver = {
	1, /* buswidth */
	N_("SST Standard Command Set"),
	N_("supported: SST 39VF020; 1x8 Bit"),
	sst_flash_autodetect8,
	sst_flash_print_info,
	sst_flash_erase_block,
	sst_flash_unlock_block,
	sst_flash_program,
	sst_flash_read_array,
};
