/*
 * $Id: pacman.c,v 1.1.1.1 2006/10/16 15:47:29 danov Exp $
 *
 * Copyright (C) 2002 ETC s.r.o.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Written by Christian Pellegrin <chri@ascensit.com>, 2003.
 * Modified by Marcel Telka <marcel@telka.sk>, 2003.
 *
 */

#include "sysdep.h"

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "part.h"
#include "bus.h"
#include "chain.h"
#include "bssignal.h"
#include "jtag.h"
#include "buses.h"

typedef struct {
	chain_t *chain;
	part_t *part;
	signal_t *cs;
	signal_t *addr[19];
	signal_t *data[8];
	signal_t *wr;
	signal_t *rd;
	signal_t *rst;
} bus_params_t;

#define	CHAIN	((bus_params_t *) bus->params)->chain
#define	PART	((bus_params_t *) bus->params)->part
#define	RCE	((bus_params_t *) bus->params)->cs
#define	RA	((bus_params_t *) bus->params)->addr
#define	RD	((bus_params_t *) bus->params)->data
#define	RWE	((bus_params_t *) bus->params)->wr
#define	ROE	((bus_params_t *) bus->params)->rd
#define	RESET	((bus_params_t *) bus->params)->rst

static void
select_flash( bus_t *bus )
{
	part_t *p = PART;

	part_set_signal( p, RCE, 1, 0 );
	part_set_signal( p, RESET, 1, 1 );
}

static void
unselect_flash( bus_t *bus )
{
	part_t *p = PART;

	part_set_signal( p, RCE, 1, 1 );
	part_set_signal( p, RESET, 1, 1 );
}

static void
setup_address( bus_t *bus, uint32_t a )
{
	int i;
	part_t *p = PART;

	for (i = 0; i < 19; i++)
		part_set_signal( p, RA[i], 1, (a >> i) & 1 );
}

static void
set_data_in( bus_t *bus )
{
	int i;
	part_t *p = PART;

	for (i = 0; i < 8; i++)
		part_set_signal( p, RD[i], 0, 0 );
}

static void
setup_data( bus_t *bus, uint32_t d )
{
	int i;
	part_t *p = PART;

	for (i = 0; i < 8; i++)
		part_set_signal( p, RD[i], 1, (d >> i) & 1 );

}

static void
pacman_bus_printinfo( bus_t *bus )
{
	int i;

	for (i = 0; i < CHAIN->parts->len; i++)
		if (PART == CHAIN->parts->parts[i])
			break;
	printf( _("Altera EP1C3 PacMan bus driver via BSR (JTAG part No. %d)\n"), i );
}

static void
pacman_bus_prepare( bus_t *bus )
{
	part_set_instruction( PART, "EXTEST" );
	chain_shift_instructions( CHAIN );
}

static void
pacman_bus_read_start( bus_t *bus, uint32_t adr )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;

	select_flash( bus );
	part_set_signal( p, ROE, 1, 0 );
	part_set_signal( p, RWE, 1, 1 );

	setup_address( bus, adr );
	set_data_in( bus );

	chain_shift_data_registers( chain, 0 );
}

static uint32_t
pacman_bus_read_next( bus_t *bus, uint32_t adr )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;
	int i;
	uint32_t d = 0;

	setup_address( bus, adr );
	chain_shift_data_registers( chain, 1 );

	for (i = 0; i < 8; i++)
		d |= (uint32_t) (part_get_signal( p, RD[i] ) << i);

	return d;
}

static uint32_t
pacman_bus_read_end( bus_t *bus )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;
	int i;
	uint32_t d = 0;

	unselect_flash( bus );
	part_set_signal( p, ROE, 1, 1 );
	part_set_signal( p, RWE, 1, 1 );

	chain_shift_data_registers( chain, 1 );

	for (i = 0; i < 8; i++)
		d |= (uint32_t) (part_get_signal( p, RD[i] ) << i);

	return d;
}

static uint32_t
pacman_bus_read( bus_t *bus, uint32_t adr )
{
	pacman_bus_read_start( bus, adr );
	return pacman_bus_read_end( bus );
}

static void
pacman_bus_write( bus_t *bus, uint32_t adr, uint32_t data )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;

	select_flash( bus );
	part_set_signal( p, ROE, 1, 1 );

	setup_address( bus, adr );
	setup_data( bus, data );

	chain_shift_data_registers( chain, 0 );

	part_set_signal( p, RWE, 1, 0 );
	chain_shift_data_registers( chain, 0 );
	part_set_signal( p, RWE, 1, 1 );
	unselect_flash( bus );
	chain_shift_data_registers( chain, 0 );
}

static int
pacman_bus_area( bus_t *bus, uint32_t adr, bus_area_t *area )
{
	area->description = NULL;
	area->start = UINT32_C(0x00000000);
	area->length = UINT64_C(0x100000000);
	area->width = 8;

	return 0;
}

static void
pacman_bus_free( bus_t *bus )
{
	free( bus->params );
	free( bus );
}

static bus_t *pacman_bus_new( void );

const bus_driver_t pacman_bus = {
	"pacman",
	N_("EP1C3 PacMan board bus driver"),
	pacman_bus_new,
	pacman_bus_free,
	pacman_bus_printinfo,
	pacman_bus_prepare,
	pacman_bus_area,
	pacman_bus_read_start,
	pacman_bus_read_next,
	pacman_bus_read_end,
	pacman_bus_read,
	pacman_bus_write
	,0,0,0,0
};

static bus_t *
pacman_bus_new( void )
{
	bus_t *bus;
	char buff[15];
	int i;
	int failed = 0;

	if (!chain || !chain->parts || chain->parts->len <= chain->active_part || chain->active_part < 0)
		return NULL;

	bus = malloc( sizeof (bus_t) );
	if (!bus)
		return NULL;

	bus->driver = &pacman_bus;
	bus->params = malloc( sizeof (bus_params_t) );
	if (!bus->params) {
		free( bus );
		return NULL;
	}

	CHAIN = chain;
	PART = chain->parts->parts[chain->active_part];

	RCE = part_find_signal( PART, "RCE" );
	if (!RCE) {
		printf( _("signal '%s' not found\n"), "RCE" );
		failed = 1;
	}

	for (i = 0; i < 19; i++) {
		sprintf( buff, "RA%d", i );
		RA[i] = part_find_signal( PART, buff );
		if (!RA[i]) {
			printf( _("signal '%s' not found\n"), buff );
			failed = 1;
			break;
		}
	}
	for (i = 0; i < 8; i++) {
		sprintf( buff, "RD%d", i);
		RD[i] = part_find_signal( PART, buff );
		if (!RD[i]) {
			printf( _("signal '%s' not found\n"), buff );
			failed = 1;
			break;
		}
	}
	RWE = part_find_signal( PART, "RWE" );
	if (!RWE) {
		printf( _("signal '%s' not found\n"), "RWE" );
		failed = 1;
	}
	ROE = part_find_signal( PART, "ROE" );
	if (!ROE) {
		printf( _("signal '%s' not found\n"), "ROE" );
		failed = 1;
	}

	RESET = part_find_signal( PART, "RESET" );
	if (!RESET) {
		printf( _("signal '%s' not found\n"), "RESET" );
		failed = 1;
	}

	if (failed) {
		free( bus->params );
		free( bus );
		return NULL;
	}

	return bus;
}
